"""
Server module, contains core-program classes.
"""
from .application.abstract_backend import AbstractBackend
from .application.standard_json_encoder import StandardJsonEncoder
from .data.uifb import UIFB
from .presentation.endpoints import Endpoints

__all__ = [
    Endpoints.__name__,
    StandardJsonEncoder.__name__,
    AbstractBackend.__name__,
    UIFB.__name__,
]



# pylint: disable=R0801
"""
Module that contains the ClingraphBackend.
"""
import textwrap
from pathlib import Path

from clingo import Control
from clingo.symbol import Function, String
from clingraph import Factbase, compute_graphs, render
from clingraph.clingo_utils import ClingraphContext
from clorm import Raw

from clinguin.server.application.backends.clingo_multishot_backend import (
    ClingoMultishotBackend,
)
from clinguin.server.data.attribute import AttributeDao

# Self defined
from clinguin.utils import StandardTextProcessing, image_to_b64


class ClingraphBackend(ClingoMultishotBackend):
    """
    Extends ClingoMultishotBackend. With this Backend it is possible to include clingraph images in the UI.
    The image is rendered based on a visualization encoding every time the UI is updated.
    Then, they are sent the client as Base64 encoding.
    """

    def __init__(self, args):
        super().__init__(args)

        self._clingraph_files = args.clingraph_files
        self._select_model = args.select_model
        self._type = args.type
        self._select_graph = args.select_graph
        self._dir = args.dir
        self._name_format = args.name_format
        self._engine = args.engine
        self._disable_saved_to_file = args.disable_saved_to_file

        self._intermediate_format = args.intermediate_format
        self._attribute_image_key = "image_type"
        self._attribute_image_value = "clingraph"

    # ---------------------------------------------
    # Overwrite
    # ---------------------------------------------

    def _update_uifb_ui(self):
        """
        Updates the ui state with the previously computed domain state.
        It uses the same domain state as input to the visualization encoding to
        create clingraph images.
        Any image is replaced by a b64 representation
        """
        super()._update_uifb_ui()
        if self._uifb.is_unsat:
            return
        graphs = self._compute_clingraph_graphs(self._uifb.conseq_facts)
        if not self._disable_saved_to_file:
            self._save_clingraph_graphs_to_file(graphs)

        self._replace_uifb_with_b64_images_clingraph(graphs)

    @classmethod
    def register_options(cls, parser):
        """
        Registers command line options for ClingraphBackend.
        """
        ClingoMultishotBackend.register_options(parser)

        parser.add_argument(
            "--clingraph-files",
            help=textwrap.dedent(
                """\
                            A visualization encoding that will be used to generate the graph facts
                            by calling clingo with the input.
                            This encoding is expected to have only one stable model."""
            ),
            # type=argparse.FileType('r'),
            nargs="+",
            metavar="",
        )

        parser.add_argument(
            "--prefix",
            default="",
            help=textwrap.dedent(
                """\
                            Prefix expected in all the considered facts.
                            Example: --prefix=viz_ will look for predicates named viz_node, viz_edge etc."""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--default-graph",
            default="default",
            help=textwrap.dedent(
                """\
                        The name of the default graph.
                        All nodes and edges with arity 1 will be assigned to this graph
                            (default: %(default)s)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--select-graph",
            help=textwrap.dedent(
                """\
                    Select one of the graphs by name.
                    Can appear multiple times to select multiple graphs"""
            ),
            type=str,
            action="append",
            nargs="?",
            metavar="",
        )

        parser.add_argument(
            "--select-model",
            help=textwrap.dedent(
                """\
                    Select only one of the models when using a json input.
                    Defined by a number starting in index 0.
                    Can appear multiple times to select multiple models."""
            ),
            type=int,
            action="append",
            nargs="?",
            metavar="",
        )

        parser.add_argument(
            "--name-format",
            help=textwrap.dedent(
                """\
                        An optional string to format the name when saving multiple files,
                        this string can reference parameters `{graph_name}` and `{model_number}`.
                        Example `new_version-{graph_name}-{model_number}`
                        (default: `{graph_name}` or
                                `{model_number}/{graph_name}` for multi model functionality from json)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--dir",
            default="out",
            help=textwrap.dedent(
                """\
                            Directory for writing the output files
                                (default: %(default)s)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--type",
            default="graph",
            choices=["graph", "digraph"],
            help=textwrap.dedent(
                """\
                        The type of the graph
                        {graph|digraph}
                            (default: %(default)s)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--engine",
            default="dot",
            choices=[
                "dot",
                "neato",
                "twopi",
                "circo",
                "fdp",
                "osage",
                "patchwork",
                "sfdp",
            ],
            help=textwrap.dedent(
                """\
                            Layout command used by graphviz
                            {dot|neato|twopi|circo|fdp|osage|patchwork|sfdp}
                                (default: %(default)s)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--disable-saved-to-file",
            action="store_true",
            help="Disable image saved to file",
        )

        parser.add_argument(
            "--intermediate-format",
            default="svg",
            type=str,
            choices=[
                "png",
                "svg",
            ],
            help="Intermediate format. Use 'svg' for angular fronted and 'png' tkinter. (default: %(default)s)",
        )

    # ---------------------------------------------
    # Private methods
    # ---------------------------------------------

    def _compute_clingraph_graphs(self, prg):
        """
        Computes all the graphs using the encoding and the domain state

        Arguments:

            prg (str): The model, brave, and cautious consequences (domain-state)
        """
        fbs = []
        ctl = Control("0")

        existant_file_counter = 0
        for f in self._clingraph_files:
            path = Path(f)
            if path.is_file():
                try:
                    ctl.load(str(f))
                    existant_file_counter += 1
                except Exception:
                    self._logger.critical(
                        "Failed to load file %s (there is likely a syntax error in this logic program file).",
                        f,
                    )
            else:
                self._logger.critical(
                    "File %s does not exist, this file is skipped.", f
                )

        if existant_file_counter == 0:
            exception_string = (
                "None of the provided clingraph files exists, but at least one syntactically"
                + "valid clingraph file must be specified. Exiting!"
            )

            self._logger.critical(exception_string)
            raise Exception(exception_string)

        ctl.add("base", [], prg)
        ctl.add("base", [], self._clinguin_state)
        ctl.ground([("base", [])], ClingraphContext())

        ctl.solve(on_model=lambda m: fbs.append(Factbase.from_model(m)))
        if self._select_model is not None:
            for m in self._select_model:
                if m >= len(fbs):
                    raise ValueError(f"Invalid model number selected {m}")
            fbs = [f if i in self._select_model else None for i, f in enumerate(fbs)]

        if len(fbs) > 1:
            self._logger.warning(
                "Multiple clingraph outputs were computed. Only first one considered."
            )
        graphs = compute_graphs([fbs[0]], graphviz_type=self._type)
        return graphs

    def _save_clingraph_graphs_to_file(self, graphs):
        """
        Saves the computed graphs in a file

        Arguments:
            graphs (dic) The computed graphs
        """
        if self._select_graph is not None:
            graphs = [
                {
                    g_name: g
                    for g_name, g in graph.items()
                    if g_name in self._select_graph
                }
                for graph in graphs
            ]
        write_arguments = {"directory": self._dir, "name_format": self._name_format}
        paths = render(
            graphs, format="png", engine=self._engine, view=False, **write_arguments
        )
        self._logger.debug("Clingraph saved images:")
        self._logger.debug(paths)

    def _replace_uifb_with_b64_images_clingraph(self, graphs):
        """
        Replaces the clingraph predicates of the UI with the computed graphs.

        Arguments:
            graphs (dic) The computed graphs
        """
        attributes = list(self._uifb.get_attributes(key=self._attribute_image_key))
        for attribute in attributes:
            attribute_value = StandardTextProcessing.parse_string_with_quotes(
                str(attribute.value)
            )
            is_cg_image = attribute_value.startswith(self._attribute_image_value)

            if not is_cg_image:
                continue

            graph_name = "default"
            split = attribute_value.split("__")
            if len(split) > 1:
                graph_name = split[1]

            # Currently assuming SVG, otherwise b64 encoding necessary!
            image_value = self._create_image_from_graph(graphs, key=graph_name)
            new_image_key = "image"
            base64_key_image = image_to_b64(image_value)

            new_attribute = AttributeDao(
                Raw(Function(str(attribute.id), [])),
                Raw(Function(str(new_image_key), [])),
                Raw(String(str(base64_key_image))),
            )
            self._uifb.add_attribute_direct(new_attribute)

            # self._uifb.replace_attribute(attribute, new_attribute)

    def _create_image_from_graph(self, graphs, position=None, key=None):
        """
        Creates the image of the graph using clingraph

        Arguments:
            graphs (dic) The computed graphs
            position (int) The position of the graph to show
            key (int) The key of the graph to show
        """
        graphs = graphs[0]
        if position is not None:
            if (len(graphs) - 1) >= position:
                graph = graphs[list(graphs.keys())[position]]
            else:
                self._logger.error("Attempted to access not valid position")
                raise Exception("Attempted to access not valid position")
        elif key is not None:
            if key in graphs:
                graph = graphs[key]
            else:
                self._logger.error("Key not found in graphs: %s", str(key))
                raise Exception("Key not found in graphs: " + str(key))
        else:
            self._logger.error("Must either specify position or key!")
            raise Exception("Must either specify position or key!")

        graph.format = self._intermediate_format
        img = graph.pipe(engine=self._engine)

        return img



# pylint: disable=R0903
"""
Module that contains the AttributeDao.
"""
from clorm import Predicate, RawField


class AttributeDao(Predicate):
    """
    Class for CLORM (clingo-object-relational-mapping), i.e. for accessing attributes in a factbase.
    """

    id = RawField
    key = RawField
    value = RawField

    class Meta:
        """
        Meta class
        """

        name = "attr"



"""
Module that contains the ClingoDL Backend.
"""


from pathlib import Path

from clingo import Control
from clingo.ast import ProgramBuilder, parse_files
from clingo.script import enable_python
from clingodl import ClingoDLTheory

from clinguin.server.application.backends.clingo_multishot_backend import (
    ClingoMultishotBackend,
)

enable_python()
# pylint: disable=attribute-defined-outside-init


class ClingoDLBackend(ClingoMultishotBackend):
    """
    Backend for explanations
    """

    # ---------------------------------------------
    # Private methods
    # ---------------------------------------------

    # ---------------------------------------------
    # Overwrite
    # ---------------------------------------------

    def _init_ctl(self):
        self._ctl = Control(["0"] + self._constants)
        self._theory = ClingoDLTheory()
        self._theory.register(self._ctl)

        existant_file_counter = 0
        with ProgramBuilder(self._ctl) as bld:
            for f in self._domain_files:
                path = Path(f)
                if path.is_file():
                    try:
                        parse_files(
                            [f], lambda ast: self._theory.rewrite_ast(ast, bld.add)
                        )
                        existant_file_counter += 1
                    except Exception:
                        self._logger.critical(
                            "Failed to load file %s (there is likely a syntax error in this logic program file).",
                            f,
                        )
                else:
                    self._logger.critical(
                        "File %s does not exist, this file is skipped.", f
                    )

        if existant_file_counter == 0:
            exception_string = (
                "None of the provided domain files exists or they can't be parsed by clingo. At least one syntactically"
                + "valid domain file must be specified."
            )
            raise Exception(exception_string)

        for atom in self._atoms:
            self._ctl.add("base", [], str(atom) + ".")

    def _solve_set_handler(self):
        # pylint: disable=attribute-defined-outside-init
        self._theory.prepare(self._ctl)
        self._handler = self._ctl.solve(
            assumptions=[(a, True) for a in self._assumptions], yield_=True
        )

    def _on_model(self, model):
        self._theory.on_model(model)
        # pylint: disable=attribute-defined-outside-init
        self._assignment = [
            f"_clinguin_assign({key},{val})."
            for key, val in self._theory.assignment(model.thread_id)
        ]

    @property
    def _clinguin_state(self):
        """
        Additional program to pass to the UI computation. It represents to the state of the backend
        """
        prg = super()._clinguin_state
        prg += " ".join(self._assignment)
        return prg

    # ---------------------------------------------
    # Plolicy methods (Overwrite ClingoMultishotBackend)
    # ---------------------------------------------



"""
Responsible for parsing the command line attributes
"""
import argparse
import importlib
import inspect
import os
import sys
import textwrap
import traceback

from .client import AbstractFrontend
from .server import AbstractBackend
from .show_frontend_syntax_enum import ShowFrontendSyntaxEnum

if sys.version_info[1] < 8:
    import importlib_metadata as metadata  # nocoverage
else:
    from importlib import metadata  # nocoverage

VERSION = metadata.version("clinguin")


class ArgumentParser:
    """
    ArgumentParser-Class, Responsible for parsing the command line attributes
    """

    default_backend_exec_string = "from .server.application.backends import *"
    default_frontend_exec_string = "from .client.presentation.frontends import *"

    default_backend = "ClingoMultishotBackend"
    default_frontend = "TkinterFrontend"

    def __init__(self) -> None:
        self.frontend_name = None
        self.frontend = None

        self.titles = {
            "client": self._client_title,
            "server": self._server_title,
            "client-server": self._client_server_title,
        }
        self.descriptions = {
            "client": "Start a client process that will render a UI.",
            "server": "Start server process making endpoints available for a client.",
            "client-server": "Start client and a server processes.",
        }
        self.backend_name = None
        self.backend = None
        self._show_frontend_syntax = ShowFrontendSyntaxEnum.NONE

    def parse(self, process, string_args):
        """
        After initialization of the ArgumentParser call this function to parse the arguments.
        """
        self._parse_custom_classes(string_args)

        parser = argparse.ArgumentParser(
            description=self._clinguin_description(process),
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )
        parser.add_argument(
            "--version", "-v", action="version", version=f"%(prog)s {VERSION}"
        )
        subparsers = parser.add_subparsers(
            title="Process type",
            description="The type of process to start: a client (UI) a server (Backend) or both",
            dest="process",
        )
        self._create_client_subparser(subparsers)
        self._create_server_subparser(subparsers)
        self._create_client_server_subparser(subparsers)

        args = parser.parse_args(string_args)

        self._add_selected_backend(args)

        return args

    def _add_selected_backend(self, args):
        args.backend = self.backend
        args.frontend = self.frontend

    @property
    def _clinguin_title(self):
        return """
              ___| (_)_ __   __ _ _   _(_)_ __
             / __| | | '_ \\ / _` | | | | | '_ \\
            | (__| | | | | | (_| | |_| | | | | |
             \\___|_|_|_| |_|\\__, |\\__,_|_|_| |_|
                            |___/
            """

    @property
    def _client_server_title(self):
        return """
             _ | o  _  ._  _|_     _  _  ._     _  ._
            (_ | | (/_ | |  |_    _> (/_ |  \\/ (/_ |

            """

    @property
    def _client_title(self):
        return """
                       _ | o  _  ._  _|_
                      (_ | | (/_ | |  |_

            """

    @property
    def _server_title(self):
        return """
                       _  _  ._     _  ._
                      _> (/_ |  \\/ (/_ |

            """

    def _clinguin_description(self, process):
        description = (
            "Clinguin is a GUI language extension for a logic program that uses Clingo."
        )
        if process not in ["server", "client", "client-server"]:
            ascci = f"{self._clinguin_title}{description}"
            return_value = f"{inspect.cleandoc(ascci)}\n\n{description}"
        else:
            ascci = f"{self._clinguin_title}{self.titles[process]}"
            return_value = f"{inspect.cleandoc(ascci)}\n\n{description}\n{self.descriptions[process]}"

        return return_value

    def _import_classes(self, path):
        if os.path.isfile(path):
            sys.path.append(os.path.dirname(path))
            self._import_files_from_path_array([path])
        else:
            sys.path.append(path)
            self._recursive_import(path, "", "")

    def _import_files_from_path_array(self, file_paths, module=""):
        for file_path in file_paths:
            base = os.path.basename(file_path)
            file_name = os.path.splitext(base)[0]
            ending = os.path.splitext(base)[1]
            if ending == ".py":
                if module != "":
                    try:
                        importlib.import_module(module + "." + file_name)
                    except Exception as ex:
                        raise Exception(
                            "Could not import module: " + module + "." + file_name
                        ) from ex
                else:
                    importlib.import_module(file_name)

    def _recursive_import(self, full_path, rec_path, module):
        folder_paths = []
        file_paths = []

        try:
            for entity in os.scandir(os.path.join(full_path, rec_path)):
                if entity.is_dir():
                    folder_paths.append(entity.path)
                elif entity.is_file():
                    file_paths.append(entity.path)
        except Exception as ex:
            print("<<<BEGIN-STACK-TRACE>>>")
            traceback.print_exc()
            print("<<<END-STACK-TRACE>>>")
            raise Exception(
                "Could not find path for importing libraries: "
                + os.path.join(full_path, rec_path)
                + ". Therefore program is terminating now (full stacktrace is printed below)."
            ) from ex

        self._import_files_from_path_array(file_paths)

        for folder_path in folder_paths:
            base = os.path.basename(folder_path)
            new_module = module
            if new_module != "":
                new_module = new_module + "." + base
            else:
                new_module = base

            self._recursive_import(full_path, os.path.join(rec_path, base), new_module)

    def _parse_custom_classes(self, str_args):
        custom_imports_parser = argparse.ArgumentParser(add_help=False)
        self._add_default_arguments_to_backend_parser(custom_imports_parser)
        self._add_default_arguments_to_client_parser(custom_imports_parser)

        args, _ = custom_imports_parser.parse_known_args(str_args)

        self.frontend_name = args.frontend
        self.backend_name = args.backend
        if args.custom_classes:
            self._import_classes(args.custom_classes)

        exec(ArgumentParser.default_backend_exec_string)
        exec(ArgumentParser.default_frontend_exec_string)

        if args.frontend_syntax and not args.frontend_syntax_full:
            self._show_frontend_syntax = ShowFrontendSyntaxEnum.SHOW
        elif args.frontend_syntax_full:
            self._show_frontend_syntax = ShowFrontendSyntaxEnum.FULL

    def _create_client_subparser(self, subparsers):
        parser_client = subparsers.add_parser(
            "client",
            help=self.descriptions["client"],
            description=self._clinguin_description("client"),
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        self._add_log_arguments(
            parser_client, abbrevation="C", logger_name="clinguin_client"
        )

        parser_client.add_argument(
            "--server-port",
            type=int,
            default=8000,
            help="Define the Port of the server, default value is 8000.",
        )

        parser_client.add_argument(
            "--server-url",
            type=str,
            default="http://localhost",
            help="Define the URL of the server, default value is 'http://localhost'.",
        )

        self._add_default_arguments_to_client_parser(parser_client)
        self.frontend = self._select_subclass_and_add_custom_arguments(
            parser_client,
            AbstractFrontend,
            self.frontend_name,
            ArgumentParser.default_frontend,
        )

        return parser_client

    def _create_server_subparser(self, subparsers):
        parser_server = subparsers.add_parser(
            "server",
            help=self.descriptions["server"],
            description=self._clinguin_description("server"),
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        self._add_log_arguments(
            parser_server, abbrevation="S", logger_name="clinguin_server"
        )
        self._add_default_arguments_to_backend_parser(parser_server)
        self.backend = self._select_subclass_and_add_custom_arguments(
            parser_server,
            AbstractBackend,
            self.backend_name,
            ArgumentParser.default_backend,
        )

        return parser_server

    def _create_client_server_subparser(self, subparsers):
        parser_server_client = subparsers.add_parser(
            "client-server",
            help=self.descriptions["client-server"],
            description=self._clinguin_description("client-server"),
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        self._add_log_arguments(
            parser_server_client,
            abbrevation="C",
            logger_name="clinguin_client",
            display_name="client-",
        )
        self._add_log_arguments(
            parser_server_client,
            abbrevation="S",
            logger_name="clinguin_server",
            display_name="server-",
        )

        self._add_default_arguments_to_client_parser(parser_server_client)
        self.frontend = self._select_subclass_and_add_custom_arguments(
            parser_server_client,
            AbstractFrontend,
            self.frontend_name,
            ArgumentParser.default_frontend,
        )

        self._add_default_arguments_to_backend_parser(parser_server_client)
        self.backend = self._select_subclass_and_add_custom_arguments(
            parser_server_client,
            AbstractBackend,
            self.backend_name,
            ArgumentParser.default_backend,
        )

        return parser_server_client

    def _add_default_arguments_to_backend_parser(self, parser):
        sub_classes = self._get_sub_classes(AbstractBackend)
        sub_class_as_options = "|".join([s.__name__ for s in sub_classes])
        sub_classes_str = "=>  Available options: {" + sub_class_as_options + "}"
        parser.add_argument(
            "--backend",
            type=str,
            help=textwrap.dedent(
                f"""\
                Optionally specify which backend to use using the class name.
                {sub_classes_str}
                """
            ),
            metavar="",
        )
        parser.add_argument(
            "--custom-classes", type=str, help="Path to custom classes.", metavar=""
        )
        parser.add_argument(
            "--server-port",
            type=int,
            default=8000,
            help="Define the server side port, default value is 8000.",
        )

    def _add_default_arguments_to_client_parser(self, parser):
        sub_classes = self._get_sub_classes(AbstractFrontend)
        sub_class_as_options = "|".join([s.__name__ for s in sub_classes])
        sub_classes_str = "=>  Available options: {" + sub_class_as_options + "}"
        parser.add_argument(
            "--frontend",
            type=str,
            help=textwrap.dedent(
                f"""\
                Optionally specify which frontend to use using the class name.
                {sub_classes_str}
                """
            ),
            metavar="",
        )
        parser.add_argument(
            "--frontend-syntax",
            action="store_true",
            help="Show available commands for the GUI.",
        )
        parser.add_argument(
            "--frontend-syntax-full",
            action="store_true",
            help="Show available commands for the GUI and shows available value-types.",
        )

    def _add_log_arguments(
        self, parser, abbrevation="", logger_name="", display_name=""
    ):
        group = parser.add_argument_group(display_name + "logger")
        group.add_argument(
            "--" + display_name + "log-disable-shell",
            action="store_true",
            help="Disable shell logging",
        )
        group.add_argument(
            "--" + display_name + "log-enable-file",
            action="store_true",
            help="Disable file logging",
        )
        group.add_argument(
            "--" + display_name + "logger-name",
            type=str,
            help="Set logger name",
            metavar="",
            default=logger_name,
        )
        group.add_argument(
            "--" + display_name + "log-level",
            type=str,
            help="Log level",
            metavar="",
            choices=["DEBUG", "INFO", "ERROR", "WARNING"],
            default="INFO",
        )
        group.add_argument(
            "--" + display_name + "log-format-shell",
            type=str,
            help="Log format shell",
            metavar="",
            default="[" + str(abbrevation) + "] %(levelname)s: %(message)s",
        )
        group.add_argument(
            "--" + display_name + "log-format-file",
            type=str,
            help="Log format file",
            metavar="",
            default="%(levelname)s: %(message)s",
        )

    def _get_sub_classes(self, cur_class):
        sub_classes = cur_class.__subclasses__()
        recursive = []
        for sub_class in sub_classes:
            recursive.extend(self._get_sub_classes(sub_class))

        return sub_classes + recursive

    def _select_subclass_and_add_custom_arguments(
        self, parser, parent, class_name, default_class
    ):
        sub_classes = self._get_sub_classes(parent)

        selected_class = None

        for sub_class in sub_classes:
            full_class_name = sub_class.__name__
            selected_by_default = not class_name and full_class_name == default_class
            selected = full_class_name == class_name
            if selected_by_default or selected:
                group = parser.add_argument_group(full_class_name)
                sub_class.register_options(group)

                should_show_frontend_syntax = self._show_frontend_syntax in [
                    ShowFrontendSyntaxEnum.SHOW,
                    ShowFrontendSyntaxEnum.FULL,
                ]

                if should_show_frontend_syntax and hasattr(
                    sub_class, "available_syntax"
                ):
                    print(sub_class.available_syntax(self._show_frontend_syntax))
                    sys.exit()

                selected_class = sub_class
        return selected_class



"""
Module that contains the Explanation Backend.
"""

import textwrap

from clingo.script import enable_python

from clinguin.server.application.backends.clingo_multishot_backend import (
    ClingoMultishotBackend,
)

enable_python()


class ExplanationBackend(ClingoMultishotBackend):
    """
    Extends ClingoMultishotBackend. This backend will treat an UNSAT result by adding the
    Minimal Unsatisfiable Core (MUC) to the domain-state, thus allowing the UI to show
    the faulty assumptions.
    """

    def __init__(self, args):
        self._muc = None
        self._lit2symbol = {}
        self._mc_base_assumptions = set()
        self._last_prg = None
        super().__init__(args)
        for a in args.assumption_signature:
            try:
                name = a.split(",")[0]
                arity = int(a.split(",")[1])
            except Exception as ex:
                raise ValueError(
                    "Argument assumption_signature must have format name,arity"
                ) from ex
            for s in self._ctl.symbolic_atoms:
                if s.symbol.match(name, arity):
                    self._mc_base_assumptions.add(s.symbol)
                    self._add_symbol_to_dict(s.symbol)
        self._assumptions = self._mc_base_assumptions.copy()

    # ---------------------------------------------
    # Private methods
    # ---------------------------------------------

    def _add_symbol_to_dict(self, symbol):
        """
        Adds a list of symbols to the mapping of symbols to literals
        """
        lit = self._ctl.symbolic_atoms[symbol].literal
        self._lit2symbol[lit] = symbol

    def _solve_core(self, assumptions):
        """
        Solves and gets the core with the basic faulty assumptions.

        Arguments:

            assumptions (list[int]): List of assumption literals
        """
        with self._ctl.solve(
            assumptions=[(a, True) for a in assumptions], yield_=True
        ) as solve_handle:
            satisfiable = solve_handle.get().satisfiable
            core = [self._lit2symbol[s] for s in solve_handle.core() if s != -1]
        return satisfiable, core

    def _get_minimum_uc(self, different_assumptions):
        """
        Computes the MUC from the assumptions

        Arguments:

            different_assumptions (list[int]): List of assumption literals that is being minimized
        """
        sat, _ = self._solve_core(assumptions=different_assumptions)

        if sat:
            return []

        assumption_set = different_assumptions
        probe_set = []

        for i, assumption in enumerate(assumption_set):
            working_set = assumption_set[i + 1 :]
            sat, _ = self._solve_core(assumptions=working_set + probe_set)
            if sat:
                probe_set.append(assumption)

                if not self._solve_core(assumptions=probe_set)[0]:
                    break

        return probe_set

    @property
    def _clinguin_state(self):
        """
        Creates the atoms that will be part of the clinguin state, which is passed to the UI computation.
        Enhances the ClingoMultishotBackend state with predicate `_muc/1` with the assumptions in the UNSAT CORE
        """
        prg = super()._clinguin_state
        if self._uifb.is_unsat:
            self._logger.info("UNSAT Answer, will add explanation")
            clingo_core = self._uifb.get_unsat_core()
            clingo_core_symbols = [self._lit2symbol[s] for s in clingo_core if s != -1]
            muc_core = self._get_minimum_uc(clingo_core_symbols)
            for s in muc_core:
                prg = prg + f"_muc({str(s)})."
        return prg

    # ---------------------------------------------
    # Overwrite
    # ---------------------------------------------

    def _add_assumption(self, predicate_symbol):
        """
        Adds an assumption by also including in the mapping to literals.
        """
        self._add_symbol_to_dict(predicate_symbol)
        self._assumptions.add(predicate_symbol)

    # ---------------------------------------------
    # Plolicy methods (Overwrite ClingoMultishotBackend)
    # ---------------------------------------------

    @classmethod
    def register_options(cls, parser):
        """
        Registers command line options for ClingraphBackend.
        """
        ClingoMultishotBackend.register_options(parser)

        parser.add_argument(
            "--assumption-signature",
            help=textwrap.dedent(
                """\
                            Signatures that will be considered as assumtions. Must be have format name,arity"""
            ),
            nargs="+",
            metavar="",
        )

    def clear_assumptions(self):
        """
        Removes all assumptions.
        """
        self._end_browsing()
        self._assumptions = self._mc_base_assumptions.copy()
        self._update_uifb()



# pylint: disable=R0903
"""
Module that contains the CallbackDao.
"""
from clorm import Predicate, RawField


class WhenDao(Predicate):
    """
    Class for CLORM (clingo-object-relational-mapping), i.e. for accessing 'when' in a factbase.
    """

    id = RawField
    event = RawField
    interaction_type = RawField
    policy = RawField

    class Meta:
        """
        Meta class
        """

        name = "when"



# pylint: disable=R0904
"""
Module that contains the UIFB class.
"""
import logging
from pathlib import Path

import clorm
from clingo import Control
from clingo.symbol import Function, Number, String
from clingraph.clingo_utils import ClingraphContext
from clorm import Raw

from clinguin.utils import Logger, NoModelError

from .attribute import AttributeDao
from .callback import WhenDao
from .element import ElementDao


class UIFB:
    """
    The UIFB is the low-level-access-class for handling clorm and clingo,
    regarding brave-cautious and other default things.
    This class provides functionality to create a factbase with brave-cautious extended files,
    functionality to query important things for clinguin, etc.
    """

    unifiers = [ElementDao, AttributeDao, WhenDao]

    def __init__(
        self,
        ui_files,
        constants,
        cautious_tag="_c",
        brave_tag="_b",
        auto_tag=None,
        include_unsat_msg=True,
    ):
        self._logger = logging.getLogger(Logger.server_logger_name)
        self._ui_files = [] if ui_files is None else ui_files
        self._tags = {"cautious": cautious_tag, "brave": brave_tag, "auto": auto_tag}
        self._conseq = {"cautious": None, "brave": None, "auto": None}
        self._factbase = None
        self._include_unsat_msg = include_unsat_msg
        self._unsat_core = None
        self._constants = constants

    def __str__(self):
        s = "\nConsequences:\n==========\n"
        for k, v in self._conseq.items():
            s += f"{k} consequences: (tagged with {self._tags[k]})\n"
            s += " ".join([str(s) for s in v])  # pylint: disable=E1133
            s += "\n"
        s += "\nUI Factbase:\n=========\n"
        s += self._factbase.asp_str()
        return s

    @property
    def is_empty(self):
        """
        Checks whether the factbase is empty.
        if so return true, else false.
        """
        return self._factbase is None

    @property
    def is_unsat(self):
        """
        Checks for unsat.
        Returns boolean.
        """
        return self._unsat_core is not None

    def tag(self, symbols, tag):
        """
        TODO -> Documentation
        """
        if tag is None:
            return symbols
        tagged = []
        for s in symbols:
            tagged.append(Function(tag, [s]))
        return tagged

    @property
    def conseq_facts(self):
        """
        TODO -> Documentation
        """
        conseq_facts = ""
        for c_type, symbols in self._conseq.items():
            if symbols is None:
                # raise RuntimeError("Use update function before getting the program")
                self._logger.warning(
                    "No %s consequences were calculated. Update consequences before updating the UI.",
                    c_type,
                )
                continue
            conseq_facts += UIFB.symbols_to_facts(self.tag(symbols, self._tags[c_type]))

        return conseq_facts

    def _set_fb_symbols(self, symbols):
        self._factbase = clorm.unify(self.unifiers, symbols)

    def ui_control(self, extra_ui_prg=""):
        """
        Generates a ClingoControl Object from paths of ui files and extra parts of a logic program given by a string.
        """
        uictl = Control(["0", "--warn=none"] + self._constants)

        existant_file_counter = 0
        for f in self._ui_files:
            path = Path(f)
            if path.is_file():
                try:
                    uictl.load(str(f))
                    existant_file_counter += 1
                except Exception as e:
                    self._logger.critical(
                        "Failed to load file %s (there is likely a syntax error in this logic program file).",
                        f,
                    )
                    self._logger.critical(str(e))
                    raise e
            else:
                self._logger.critical("File %s does not exist", f)
                raise Exception(f"File {f} does not exist")

        if existant_file_counter == 0:
            exception_string = "None of the provided ui files exists"
            self._logger.warning(exception_string)

        if self._include_unsat_msg:
            uictl.add("base", [], UIFB.get_unsat_messages_ui_encoding())

        uictl.add("base", [], extra_ui_prg)
        uictl.add("base", [], self.conseq_facts)
        uictl.add("base", [], "#show elem/3. #show attr/3. #show when/4.")
        uictl.ground([("base", [])], ClingraphContext)

        return uictl

    def from_ctl(self, ctl):
        """
        Solves a control object and sets the result of the answer sets as the factbase symbols.
        """
        with ctl.solve(yield_=True) as result:
            for m in result:
                model_symbols = m.symbols(shown=True, atoms=True)
                break

        return self._set_fb_symbols(symbols=model_symbols)

    def set_auto_conseq(self, model, on_model=lambda m: None):
        """
        Sets the auto conseq.
        Arguments:
            on_model: Optional callback to edit the models (Used for theory extensions)
        """
        on_model(model)
        model_symbols = model.symbols(shown=True, atoms=True)
        self._conseq["auto"] = model_symbols

    def get_auto_conseq(self):
        """
        Gets the auto conseq.
        """
        return self._conseq["auto"]

    def update_all_consequences(self, ctl, assumptions=None, on_model=lambda m: None):
        """
        Updates all consequences.
        Arguments:
            on_model: Optional callback to edit the models (Used for theory extensions)
        """
        c_types = ["brave", "cautious", "auto"]
        for c_type in c_types:
            try:
                self.update_consequence(c_type, ctl, assumptions, on_model=on_model)
            except NoModelError:
                # Error should be handled in the ui encoding
                return

    def update_ui(self, extra_ui_prg=""):
        """
        Computes the answer sets for the ui and adds the resulting facts accordingly.
        """
        self._logger.debug("Computing Ui with additional program:\n%s", extra_ui_prg)
        uictl = self.ui_control(extra_ui_prg)

        with uictl.solve(yield_=True) as result:
            for m in result:
                model_symbols = m.symbols(shown=True, atoms=True)
                # print(model_symbols)
                break

        self._factbase = clorm.unify(self.__class__.unifiers, model_symbols)

    def _compute_consequences(self, ctl, assumptions, on_model=lambda m: None):
        with ctl.solve(
            assumptions=[(a, True) for a in assumptions], yield_=True
        ) as result:
            model_symbols = None
            for m in result:
                on_model(m)
                model_symbols = m.symbols(shown=True, atoms=True)

            if model_symbols is None:
                self._logger.warning(
                    "Got an UNSAT result with the given domain encoding."
                )
                self._unsat_core = result.core()
                raise NoModelError()

            self._unsat_core = None

        return list(model_symbols)

    def update_consequence(
        self, c_type, ctl, assumptions=None, on_model=lambda m: None
    ):
        """
        Computes for one consequence (brave/cautious/other) all consequences.
        Arguments:
            on_model: Optional callback to edit the models (Used for theory extensions)
        """
        self._logger.debug("Updating %s consequences", c_type)
        if c_type in ["brave", "cautious"]:
            ctl.configuration.solve.models = 0
            ctl.configuration.solve.opt_mode = "ignore"
        else:
            ctl.configuration.solve.models = 1
            ctl.configuration.solve.opt_mode = "ignore"
        ctl.configuration.solve.enum_mode = c_type
        self._conseq[c_type] = self._compute_consequences(
            ctl, assumptions, on_model=on_model
        )

    def add_message(self, title, message, attribute_type="info"):
        """
        Adds a ''Message'' (aka. Notification/Pop-Up) for the user with a certain title and message.
        """
        mid = f"{hash(message)}"
        self.add_element(mid, "message", "window")
        self.add_attribute(mid, "title", title)
        self.add_attribute(mid, "message", message)
        self.add_attribute(mid, "type", attribute_type)

    # Manage factbase

    def add_element(self, cid, t, parent):
        """
        Adds an element to the factbase.
        """
        if isinstance(cid, str):
            cid = Function(cid, [])
        if isinstance(t, str):
            t = Function(t, [])
        if isinstance(parent, str):
            parent = Function(parent, [])
        self._factbase.add(ElementDao(Raw(cid), Raw(t), Raw(parent)))

    def add_attribute(self, cid, key, value):
        """
        Adds an attribute to the factbase.
        """
        if isinstance(cid, str):
            cid = Function(cid, [])
        if isinstance(key, str):
            key = Function(key, [])
        if isinstance(value, str):
            value = String(value)
        if isinstance(value, int):
            value = Number(value)
        self._factbase.add(AttributeDao(Raw(cid), Raw(key), Raw(value)))

    def add_attribute_direct(self, new_attribute):
        """
        Directly adds an attribute.
        """
        self._factbase.add(new_attribute)

    def get_elements(self):
        """
        Get all elements.
        """
        return self._factbase.query(ElementDao).all()

    def get_attributes(self, key=None):
        """
        Get all attributes.
        """
        q = self._factbase.query(AttributeDao)
        if key is not None:
            q.where(AttributeDao.key == key)
        return q.all()

    def get_callbacks(self):
        """
        Get all callbacks.
        """
        return self._factbase.query(WhenDao).all()

    def get_attributes_grouped(self):
        """
        Get all attributes grouped by element id.
        """
        return self._factbase.query(AttributeDao).group_by(AttributeDao.id).all()

    def get_callbacks_grouped(self):
        """
        Get all callbacks grouped by element id.
        """
        return self._factbase.query(WhenDao).group_by(WhenDao.id).all()

    def get_attributes_for_element_id(self, element_id):
        """
        Get all attributes for one element id.
        """
        return (
            self._factbase.query(AttributeDao)
            .where(AttributeDao.id == element_id)
            .all()
        )

    def get_callbacks_for_element_id(self, element_id):
        """
        Get all callbacks for one element id.
        """
        return self._factbase.query(WhenDao).where(WhenDao.id == element_id).all()

    def replace_attribute(self, old_attribute, new_attribute):
        """
        Replaces the old_attribute with the new_attribute.
        """
        self._factbase.remove(old_attribute)
        self._factbase.add(new_attribute)

    def get_unsat_core(self):
        """
        Gets the unsat core variable.
        """
        return self._unsat_core

    @classmethod
    def symbols_to_facts(cls, symbols):
        """
        Converts a iterable symbols to a string of facts.
        """
        return "\n".join([str(s) + "." for s in symbols])

    @classmethod
    def get_unsat_messages_ui_encoding(cls):
        """
        Get the standard unsat encoding for error messages.
        """
        return """
element(message_unsat,message,window):-_clinguin_unsat.
attribute(message_unsat,title,"Error"):-_clinguin_unsat.
attribute(message_unsat,message,"Unsatisfiable output."):-_clinguin_unsat.
attribute(message_unsat,type,error):-_clinguin_unsat.
"""



# pylint: disable=R0801
"""
Module that contains the ClingoMultishotBackend.
"""

from clingo import parse_term
from clingo.script import enable_python

from clinguin.server.application.backends import ClingoBackend

enable_python()


class ClingoMultishotBackend(ClingoBackend):
    """
    Extends the basic clingo functionality for grounding and solving with the use of assumptions and externals.

    It is selected as the default Backend
    """

    # ---------------------------------------------
    # Required methods
    # ---------------------------------------------

    # ---------------------------------------------
    # Private methods
    # ---------------------------------------------

    def _init_setup(self):
        """
        Initializes the arguments when the server starts or after a restart.
        These arguments include, the handler and iterator for browsing answer sets,
        as well as the domain control, the atoms, assumptions and externals
        """
        super()._init_setup()
        # To make static linters happy
        self._assumptions = set()
        self._externals = {"true": set(), "false": set(), "released": set()}

    @property
    def _clinguin_state(self):
        """
        Creates the atoms that will be part of the clinguin state, which is passed to the UI computation.

        Includes predicates  _clinguin_browsing/0, _clinguin_context/2 and _clinguin_assume/1
        """
        prg = super()._clinguin_state
        state_prg = "#defined _clinguin_assume/1."
        for a in self._assumptions:
            state_prg += f"_clinguin_assume({str(a)})."
        return prg + state_prg

    @property
    def _output_prg(self):
        """
        Generates the output program used when downloading into a file.
        Includes all assumptions as facts.
        """
        prg = super()._output_prg
        for a in self._assumptions:
            prg = prg + f"{str(a)}.\n"
        return prg

    def _solve_set_handler(self):
        """
        Sets the solver handler by calling the solve method of clingo with the selected assumptions
        """
        # pylint: disable=attribute-defined-outside-init
        self._handler = self._ctl.solve(
            assumptions=[(a, True) for a in self._assumptions], yield_=True
        )

    def _add_assumption(self, predicate_symbol):
        """
        Adds an assumption to the set
        """
        self._assumptions.add(predicate_symbol)

    def _update_uifb_consequences(self):
        """
        Updates the brave and cautious consequences of the domain state, by calling clingo with the assumptions
        """
        self._uifb.update_all_consequences(self._ctl, self._assumptions, self._on_model)
        if self._uifb.is_unsat:
            self._logger.error(
                "domain files are UNSAT. Setting _clinguin_unsat to true"
            )

    # ---------------------------------------------
    # Policies
    # ---------------------------------------------

    def clear_assumptions(self):
        """
        Removes all assumptions.
        """
        self._end_browsing()
        # pylint: disable=attribute-defined-outside-init
        self._assumptions = set()

        self._update_uifb()

    def add_assumption(self, predicate):
        """
        Adds an assumption

        Arguments:

            predicate (str): The clingo symbol to be added
        """
        predicate_symbol = parse_term(predicate)
        if predicate_symbol not in self._assumptions:
            self._add_assumption(predicate_symbol)
            self._end_browsing()
            self._update_uifb()

    def remove_assumption(self, predicate):
        """
        Removes an assumption

        Arguments:

            predicate (str): The clingo symbol to be removed
        """
        predicate_symbol = parse_term(predicate)
        if predicate_symbol in self._assumptions:
            self._assumptions.remove(predicate_symbol)
            self._end_browsing()
            self._update_uifb()

    def remove_assumption_signature(self, predicate):
        """
        Removes predicates matching the predicate description.

        Arguments:

            predicate (str): The predicate description as a symbol,
                where the reserver word `any` is used to state that anything can
                take part of that position. For instance, `person(anna,any)`,
                will remove all assumptions of predicate person, where the first argument is anna.
        """
        predicate_symbol = parse_term(predicate)
        arity = len(predicate_symbol.arguments)
        to_remove = []
        for s in self._assumptions:
            if s.match(predicate_symbol.name, arity):
                for i, a in enumerate(predicate_symbol.arguments):
                    if str(a) != "any" and s.arguments[i] != a:
                        break
                else:
                    to_remove.append(s)
                    continue
        for s in to_remove:
            self._assumptions.remove(s)
        if len(to_remove) > 0:
            self._end_browsing()
            self._update_uifb()

    def set_external(self, predicate, value):
        """
        Sets the value of an external.

        Arguments:

            predicate (str): The clingo symbol to be set
            value (str): The value (release, true or false)
        """
        symbol = parse_term(predicate)
        name = value
        self._end_browsing()

        if name == "release":
            self._ctl.release_external(parse_term(predicate))
            self._externals["released"].add(symbol)

            if symbol in self._externals["true"]:
                self._externals["true"].remove(symbol)

            if symbol in self._externals["false"]:
                self._externals["false"].remove(symbol)

        elif name == "true":
            self._ctl.assign_external(parse_term(predicate), True)
            self._externals["true"].add(symbol)

            if symbol in self._externals["false"]:
                self._externals["false"].remove(symbol)

        elif name == "false":
            self._ctl.assign_external(parse_term(predicate), False)
            self._externals["false"].add(symbol)

            if symbol in self._externals["true"]:
                self._externals["true"].remove(symbol)

        else:
            raise ValueError(
                f"Invalid external value {name}. Must be true, false or relase"
            )

        self._update_uifb()

    def select(self):
        """
        Select the current solution during browsing.
        All atoms in the solution are added as assumptions in the backend.
        """
        self._end_browsing()
        last_model_symbols = self._uifb.get_auto_conseq()
        symbols_to_ignore = self._externals["true"]
        symbols_to_ignore.union(self._externals["false"])
        for s in last_model_symbols:  # pylint: disable=E1133
            if s not in symbols_to_ignore:
                self._add_assumption(s)
        self._update_uifb()



# pylint: disable=R0903
"""
Module that contains the ElementDao.
"""
from clorm import Predicate, RawField


class ElementDao(Predicate):
    """
    Class for CLORM (clingo-object-relational-mapping), i.e. for accessing elements in a factbase.
    """

    id = RawField
    type = RawField
    parent = RawField

    class Meta:
        """
        Meta class
        """

        name = "elem"



# pylint: disable=R0801
"""
Module that contains the ClingoMultishotBackend.
"""
import os
from pathlib import Path

from clingo import Control, parse_term
from clingo.script import enable_python
from clingo.symbol import Function, String
from clorm import Raw

from clinguin.server import UIFB, AbstractBackend, StandardJsonEncoder
from clinguin.server.data.attribute import AttributeDao
from clinguin.utils import StandardTextProcessing, image_to_b64

enable_python()
# pylint: disable=attribute-defined-outside-init


class ClingoBackend(AbstractBackend):
    """
    The ClingoBackend contains the basic clingo functionality for grounding and solving.

    When started it sets up all the arguments provided via the command line,
    and creates a control object (domain-control) with the provided domain files.
    It grounds the program and creates an empty UI state.
    """

    def __init__(self, args):
        super().__init__(args)

        self._domain_files = [] if args.domain_files is None else args.domain_files
        self._ui_files = args.ui_files
        self._constants = [f"-c {v}" for v in args.const] if args.const else []

        self._init_setup()
        self._end_browsing()
        self._init_ctl()
        self._ground()

        include_unsat_msg = not args.ignore_unsat_msg
        self._uifb = UIFB(
            self._ui_files,
            self._constants,
            include_unsat_msg=include_unsat_msg,
        )

        self._encoding = "utf-8"
        self._attribute_image_key = "image"

    # ---------------------------------------------
    # Required methods
    # ---------------------------------------------

    def get(self):
        """
        Updates the UI and transforms the facts into a JSON.
        This method will be automatically called after executing all the operations.
        Thus, it needs to be implemented by all backends.

        The UI is only updated if there are any changes.
        """
        if self._uifb.is_empty:
            self._update_uifb()
        self._logger.debug(self._uifb)
        json_structure = StandardJsonEncoder.encode(self._uifb)
        return json_structure

    @classmethod
    def register_options(cls, parser):
        """
        Registers options in the command line for the domain files and ui files.
        """
        parser.add_argument(
            "--domain-files",
            nargs="+",
            help="Files with the domain specific encodings and the instances",
            metavar="",
        )
        parser.add_argument(
            "--ui-files",
            nargs="+",
            help="Files with the encodings that generate the UI predicates: elem, attr and when",
            metavar="",
        )
        parser.add_argument(
            "-c",
            "--const",
            nargs="+",
            help="Constant passed to clingo, <id>=<term> replaces term occurrences of <id> with <term>",
            metavar="",
        )
        parser.add_argument(
            "--ignore-unsat-msg",
            action="store_true",
            help="The automatic pop-up message in the UI when the domain files are UNSAT, will be ignored.",
        )

    # ---------------------------------------------
    # Private methods
    # ---------------------------------------------

    def _init_setup(self):
        """
        Initializes the arguments when the server starts or after a restart.
        These arguments include, the handler and iterator for browsing answer sets,
        as well as the domain control and the atoms.
        """
        # For browising
        self._handler = None
        self._iterator = None

        # To make static linters happy
        self._atoms = set()
        self._ctl = None

    def _init_ctl(self):
        """
        Initializes the control object (domain-control).
        It is used when the server is started or after a restart.
        Uses the provided constants and domain files.
        It adds the atoms.
        """
        self._ctl = Control(["0"] + self._constants)

        existant_file_counter = 0
        for f in self._domain_files:
            path = Path(f)
            if path.is_file():
                try:
                    self._ctl.load(str(f))
                    existant_file_counter += 1
                except Exception as e:
                    self._logger.critical(
                        "Failed to load file %s (there is likely a syntax error in this logic program file).",
                        f,
                    )
                    self._logger.critical(str(e))
                    raise e

            else:
                self._logger.critical("File %s does not exist", f)
                raise Exception(f"File {f} does not exist")

        for atom in self._atoms:
            self._ctl.add("base", [], str(atom) + ".")

    def _ground(self, program="base"):
        """
        Grounds the provided program

        Arguments:
            program (str): The name of the program to ground (defaults to "base")
        """
        self._ctl.ground([(program, [])])

    def _end_browsing(self):
        """
        Any current interation in the models wil be terminated by canceling the search and removing the iterator
        """
        if self._handler:
            self._handler.cancel()
            self._handler = None
        self._iterator = None

    @property
    def _is_browsing(self):
        """
        Checks if clinguin is in browsing mode.
        """
        return self._iterator is not None

    @property
    def _clinguin_state(self):
        """
        Creates the atoms that will be part of the clinguin state, which is passed to the UI computation.

        Includes predicates  _clinguin_browsing/0 and _clinguin_context/2
        """
        state_prg = "#defined _clinguin_browsing/0. #defined _clinguin_context/2. "
        if self._is_browsing:
            state_prg += "_clinguin_browsing."
        if self._uifb.is_unsat:
            state_prg += "_clinguin_unsat."
        for a in self.context:
            value = str(a.value)
            try:
                symbol = parse_term(value)
            except Exception:
                symbol = None
            if symbol is None:
                value = f'"{value}"'
            state_prg += f"_clinguin_context({str(a.key)},{value})."
        return state_prg

    @property
    def _output_prg(self):
        """
        Generates the output program used when downloading into a file
        """
        prg = ""
        for a in self._atoms:
            prg = prg + f"{str(a)}.\n"
        return prg

    def _on_model(self, model):
        """
        This method is called each time a model is obtained by the domain control.
        It can be used to extend the given model in Theory Solving.

        Arguments:
            model (clingo.Model): The found clingo model
        """

    def _update_uifb_consequences(self):
        """
        Updates the brave and cautious consequences of the domain state.
        """
        self._uifb.update_all_consequences(self._ctl, [], self._on_model)
        if self._uifb.is_unsat:
            self._logger.error(
                "Domain files are UNSAT. Setting _clinguin_unsat to true"
            )

    def _update_uifb_ui(self):
        """
        Updates the ui state with the previously computed domain state.
        Any image is replaced by a b64 representation
        """
        self._uifb.update_ui(self._clinguin_state)

        self._replace_uifb_with_b64_images()

    def _update_uifb(self):
        """
        Updates the domain-state (brave and cautious consequences) and the ui-state
        """
        self._update_uifb_consequences()
        self._update_uifb_ui()

    def _set_auto_conseq(self, model):
        """
        Sets the given model in the domain-state

        Arguments:
            model (clingo.Model): The model found by the solving
        """
        self._uifb.set_auto_conseq(model)

    def _solve_set_handler(self):
        """
        Sets the solver handler by calling the solve method of clingo
        """
        # pylint: disable=attribute-defined-outside-init
        self._handler = self._ctl.solve(yield_=True)

    def _add_atom(self, predicate_symbol):
        """
        Adds an atom if it hasn't been already aded

        Arguments:
            predicate_symbol (clingo.Symbool): The symbol for the atom
        """
        if predicate_symbol not in self._atoms:
            self._atoms.add(predicate_symbol)

    def _replace_uifb_with_b64_images(self):
        """
        Replaces all images in the ui-state by b64
        """
        attributes = list(self._uifb.get_attributes())
        for attribute in attributes:
            if str(attribute.key) != self._attribute_image_key:
                continue

            attribute_value = StandardTextProcessing.parse_string_with_quotes(
                str(attribute.value)
            )

            if os.path.isfile(attribute_value):
                with open(attribute_value, "rb") as image_file:
                    encoded_string = image_to_b64(image_file.read())
                    new_attribute = AttributeDao(
                        Raw(Function(str(attribute.id), [])),
                        Raw(Function(str(attribute.key), [])),
                        Raw(String(str(encoded_string))),
                    )
                    self._uifb.replace_attribute(attribute, new_attribute)

    # ---------------------------------------------
    # Policies
    # ---------------------------------------------

    def restart(self):
        """
        Restarts the backend by initializing parameters, controls, ending the browsing grounding and updating the UI
        """
        self._init_setup()
        self._end_browsing()
        self._init_ctl()
        self._ground()
        self._update_uifb()

    def update(self):
        """
        Updates the UI and transforms the output into a JSON.
        """
        self._update_uifb()

    def download(
        self, show_prg=None, file_name="clinguin_download.lp", domain_files=True
    ):
        """
        Downloads the current state of the backend. All added atoms and assumptions
        are put together as a list of facts.

        Arguments:
            show_prg (_type_, optional): Program to filter output using show statements. Defaults to None.
            file_name (str, optional): The name of the file for the download. Defaults to "clinguin_download.lp".
            domain_files (bool, optional): If the domain files should be included. Defaults to True

        """
        prg = self._output_prg
        was_browsing = self._is_browsing
        self._end_browsing()
        self._update_uifb()
        if was_browsing:
            self._uifb.add_message(
                "Warning",
                "Browsing was active during download, only selected solutions will be present on the file.",
                "warning",
            )
        if show_prg is not None:
            ctl = Control()
            ctl.add("base", [], prg)
            if domain_files:
                for f in self._domain_files:
                    ctl.load(f)
            ctl.add("base", [], show_prg.replace('"', ""))
            ctl.ground([("base", [])])
            with ctl.solve(yield_=True) as hnd:
                for m in hnd:
                    atoms = [f"{str(s)}." for s in m.symbols(shown=True)]

            prg = "\n".join(atoms)

        file_name = file_name.strip('"')
        with open(file_name, "w", encoding="UTF-8") as file:
            file.write(prg)
        self._uifb.add_message(
            "Download successful", f"Information saved in file {file_name}.", "success"
        )

    def clear_atoms(self):
        """
        Removes all atoms and resets the backend (i.e. it regrounds, etc.)
        and finally updates the model and returns the updated gui as a Json structure.
        """
        self._end_browsing()
        self._atoms = set()
        self._init_ctl()
        self._ground()

        self._update_uifb()

    def add_atom(self, predicate):
        """
        Adds an assumption, restarts the control and grounds again

        Arguments:

            predicate (str): The clingo symbol to be added
        """
        predicate_symbol = parse_term(predicate)
        if predicate_symbol not in self._atoms:
            self._add_atom(predicate_symbol)
            self._init_ctl()
            self._ground()
            self._end_browsing()
            self._update_uifb()

    def remove_atom(self, predicate):
        """
        Removes an assumption, restarts the control and grounds again

        Arguments:

            predicate (str): The clingo symbol to be added
        """
        predicate_symbol = parse_term(predicate)
        if predicate_symbol in self._atoms:
            self._atoms.remove(predicate_symbol)
            self._init_ctl()
            self._ground()
            self._end_browsing()
            self._update_uifb()

    def next_solution(self, opt_mode="ignore"):
        """
        Obtains the next solution. If a no browsing has been started yet, then it calls solve,
        otherwise it iterates the models in the last call.

        Arguments:
            opt_mode: The clingo optimization mode, bu default is 'ignore', to browse only optimal models use 'optN'
        """
        if self._ctl.configuration.solve.opt_mode != opt_mode:
            self._logger.debug("Ended browsing since opt mode changed")
            self._end_browsing()
        optimizing = opt_mode in ["optN", "opt"]
        if not self._iterator:
            self._ctl.configuration.solve.enum_mode = "auto"
            self._ctl.configuration.solve.opt_mode = opt_mode
            self._ctl.configuration.solve.models = 0
            self._solve_set_handler()
            self._iterator = iter(self._handler)
        try:
            model = next(self._iterator)
            while optimizing and not model.optimality_proven:
                self._logger.info("Skipping non-optimal model")
                model = next(self._iterator)
            self._set_auto_conseq(model)
            self._update_uifb_ui()
        except StopIteration:
            self._logger.info("No more solutions")
            self._end_browsing()
            self._update_uifb()
            self._uifb.add_message("Browsing Information", "No more solutions")

    def select(self):
        """
        Select the current solution during browsing.
        All atoms in the solution are added as atoms in the backend.
        """
        self._end_browsing()
        last_model_symbols = self._uifb.get_auto_conseq()
        for s in last_model_symbols:  # pylint: disable=E1133
            self._add_atom(s)
        self._update_uifb()



"""
Module which contains the AbstractBackend.
"""
import logging

from clinguin.utils import CustomArgs

# Like an interface


class AbstractBackend(CustomArgs):
    """
    Root class of all Backends. Sets up the logger, arguments and context.
    """

    def __init__(self, args):
        """
        Initializes the class by setting the logger arguments and context
        """
        self._logger = logging.getLogger(args.log_args["name"])
        self.args = args
        self.context = []

    def get(self):
        """
        Updates the UI and transforms the facts into a JSON.
        This method will be automatically called after executing all the operations.
        Thus, it needs to be implemented by all backends.
        """
        raise NotImplementedError("All backends must implement the method _get")

    def _set_context(self, context):
        """
        Sets the context

        Arguments:
            context: The context dictionary
        """
        self.context = context

    @classmethod
    def register_options(cls, parser):
        """
        Registers options in the command line for a given parser.
        It is automatically called on start for the provided backends.
        """
